#include <iostream>
#include <strstream>
#include <stdio.h>

#include "ko/PhysicalDolfinInterface.hpp"

namespace ko
{

PhysicalDolfinInterface::PhysicalDolfinInterface(dolfin::Mesh& mesh,
						 Physical& physical,
						 dolfin::Function& velocity,
						 dolfin::real young,
						 dolfin::real damp) :
  mesh(mesh), boundary(0), physical(physical), velocity(velocity),
  young(young), damp(damp)

{
  if(mesh.numCells() > 0)
  {
    boundary = new dolfin::BoundaryMesh(mesh, vertex_map, cell_map);
    build();
  }
  else
  {
    boundary = new dolfin::Mesh();
  }

  updateMaterial();
}

void PhysicalDolfinInterface::updatePhysical()
{
  for(dolfin::VertexIterator nmesh(*boundary); !nmesh.end(); ++nmesh)
  {
    dolfin::Vertex& vertex = *nmesh;

    dolfin::Vertex iv(mesh, vertex_map(vertex));
    
    updateVertex(iv, vertex);
  }
}

void PhysicalDolfinInterface::build()
{
  elements.resize(mesh.numVertices());
  vertices.resize(mesh.numVertices());

  //for(dolfin::VertexIterator nmesh(boundary); !nmesh.end(); ++nmesh)
  for(dolfin::VertexIterator nmesh(*boundary); !nmesh.end(); ++nmesh)
  {
    dolfin::Vertex& vertex = *nmesh;

    addVertex(vertex);
  }

  for(dolfin::CellIterator c(*boundary); !c.end(); ++c)
  {
    dolfin::Cell &cell = *c;
    
    addBoundaryCell(cell);
  }

  m.name = "material";

  physical.materials.push_back(&m);


//   Connect nearby elements together (a tetrahedral pattern is too sparse)
  for(std::list<Element *>::iterator it = physical.elements.begin();
      it != physical.elements.end(); it++)
  {
    Element &e = **it;
    
    for(std::list<Connection *>::iterator it2 = e.connections.begin();
	it2 != e.connections.end(); it2++)
    {
      Connection &ce = **it2;
      
      Element *etmp = 0;
      
      if(ce.element1 == &e)
      {
	etmp = ce.element2;
      }
      else
      {
	etmp = ce.element1;
      }
      
      for(std::list<Connection *>::iterator it3 = etmp->connections.begin();
	  it3 != etmp->connections.end(); it3++)
      {
	Connection &ce = **it3;
	
	Element *e2 = 0;
	
	if(ce.element1 == &e)
	{
	  e2 = ce.element2;
	}
	else
	{
	  e2 = ce.element1;
	}
	
	Vector r(3);
	r = e.position;
	r.axpy(-1, e2->position);
	
	dolfin::real distance;
	
	distance = r.norm();
	
	if(distance <= 3.0 * e.radius)
	{
	  if(!e.isConnected(*e2))
	  {
	    Connection &c = *(new Connection());
	    
	    c.connect(e, *e2);
	    c.length = distance;
	    physical.addConnection(c);
	    
// 	      std::cerr << "added connection" << std::endl;
	  }
	}
      }
    }
  }

  updatePhysical();
}

void PhysicalDolfinInterface::addVertex(dolfin::Vertex& vertex)
{
  Element &e = *(new Element());

  physical.elements.push_back(&e);
  elements[vertex.index()] = &e;
  vertices[vertex.index()] = &vertex;

  dolfin::Vertex iv(mesh, vertex_map(vertex));

  updateVertex(iv, vertex);

  e.material = &m;

  e.index = vertex.index();

  std::ostrstream name;
  name << vertex.index() << std::ends;
  
  e.name = name.str();

  // Determine radius
  // 0.3 * max of neighbor cell diameter

//   dolfin::real maxr = DBL_MAX;

  dolfin::TopologyComputation::computeConnectivity(mesh, 0, 2);
  dolfin::TopologyComputation::computeConnectivity(mesh, 2, 0);

  dolfin::TopologyComputation::computeConnectivity(mesh, 0, 3);
  dolfin::TopologyComputation::computeConnectivity(mesh, 3, 0);

  dolfin::real sum = 0.0;
  for(dolfin::CellIterator cell(vertex); !cell.end(); ++cell)
  {
    dolfin::Cell &c = *cell;
//     maxr = std::min(maxr, c.diameter());
    std::cout << "cell: " << c.index() << std::endl;
    std::cout << "h: " << c.diameter() << std::endl;

    sum += c.diameter();
  }
  sum /= vertex.numEntities(2);
//   maxr *= 0.3;
//   sum *= 0.40;
  sum *= 0.50;

  e.radius = sum;
}

void PhysicalDolfinInterface::addCell(dolfin::Cell &cell)
{
  for(int i = 0; i < 4; i++)
  {
    for(int j = 0; j < 4; j++)
    {
      Element &e0 = *(elements[cell.entities(0)[i]]);
      Element &e1 = *(elements[cell.entities(0)[j]]);

      if((i != j) && !e0.isConnected(e1))
      {
	Connection &c = *(new Connection());
	
	Vector r(3);
	double rnorm;
	
	r = e0.position;
	r.axpy(-1, e1.position);
	
	rnorm = r.norm();
	
	c.connect(e0, e1);
	c.length = rnorm;
	
	physical.connections.push_back(&c);
      }
    }
  }
}

void PhysicalDolfinInterface::addBoundaryCell(dolfin::Cell &cell)
{
  for(dolfin::EdgeIterator nedge(cell); !nedge.end(); ++nedge)
  {
    dolfin::Edge& edge = *nedge;

    Element &e0 = *(elements[edge.entities(0)[0]]);
    Element &e1 = *(elements[edge.entities(0)[1]]);

    if(!e0.isConnected(e1))
    {
      Connection &c = *(new Connection());
      
      Vector r(3);
      double rnorm;
      
      r = e0.position;
      r.axpy(-1, e1.position);
      
      rnorm = r.norm();
      
      c.connect(e0, e1);
      c.length = rnorm;
      
      physical.connections.push_back(&c);
    }
  }

}

void PhysicalDolfinInterface::updateVertex(dolfin::Vertex& iv,
                                           dolfin::Vertex& bv)
{
  Element &e = *(elements[bv.index()]);

//   e.position(0) = vertex.point().x();
//   e.position(1) = vertex.point().y();
//   e.position(2) = vertex.point().z();
  e.position(0) = iv.coordinates()[0];
  e.position(1) = iv.coordinates()[1];
  e.position(2) = iv.coordinates()[2];

  e.velocity(0) = velocity(iv, 0);
  e.velocity(1) = velocity(iv, 1);
  e.velocity(2) = velocity(iv, 2);
}

void PhysicalDolfinInterface::updateMaterial()
{
  m.young = 1.0 * young;
  m.damp = 0.5 * damp;
  m.friction = 0.4;
  m.density = 1e3;
  m.strength = 1e10;
}

void PhysicalDolfinInterface::computefcontact(dolfin::Vector& fcontact,
					      MechanicsModel& mmodel)
{
  updatePhysical();

  fcontact = 0.0;

  for(std::list<ko::Element *>::iterator it = physical.elements.begin();
      it != physical.elements.end(); it++)
  {
    Element *e = *it;
    int i = e->index;

    e->force = 0.0;
  }

  for(std::list<ko::Connection *>::iterator it = physical.connections.begin();
      it != physical.connections.end(); it++)
  {
    Connection &c = **it;
    if(c.transient)
    {
      static Vector force(3);
      
      Element *e1 = c.element1;
      Element *e2 = c.element2;
      
      mmodel.force(c, *e1, force);
      e1->force += force;
      force *= -1;
      e2->force += force;
      
      // Project force on connection
      static Vector r(3), tmp(3);
      double rnorm;
      
      r = e1->position;
      r.axpy(-1, e2->position);
      
      rnorm = r.norm();
      
      tmp = r;
      tmp *= (force * r) / (rnorm * rnorm);
      
      c.tension = tmp.norm();
      
      if(force * r < 0)
	c.tension *= -1;
      
//       cout << "tension: " << c.tension << endl;
    }
  }

  int offset = mesh.numVertices();
  for(dolfin::VertexIterator nmesh(*boundary); !nmesh.end(); ++nmesh)
  {
    dolfin::Vertex& vertex = *nmesh;

    Element *e = elements[vertex.index()];

    dolfin::Vertex iv(mesh, vertex_map(vertex));

//     int i = e->index;
    int i = iv.index();

    double mass = e->material->density * pow(2 * e->radius, 3);

    fcontact(0 * offset + i) = e->force(0) / mass;
    fcontact(1 * offset + i) = e->force(1) / mass;
    fcontact(2 * offset + i) = e->force(2) / mass;
  }

}

}


